"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.myQGarageDoor = void 0;
const myq_accessory_1 = require("./myq-accessory");
const settings_1 = require("./settings");
class myQGarageDoor extends myq_accessory_1.myQAccessory {
    constructor() {
        super(...arguments);
        this.batteryDeviceSupport = false;
        this.obstructionDetected = false;
    }
    // Configure a garage door accessory for HomeKit.
    configureDevice() {
        var _a;
        const accessory = this.accessory;
        const device = accessory.context.device;
        const doorInitialState = accessory.context.doorState;
        const hap = this.hap;
        // Clean out the context object.
        accessory.context = {};
        accessory.context.device = device;
        accessory.context.doorState = doorInitialState;
        const gdOpener = accessory.getService(hap.Service.GarageDoorOpener);
        // Clear out stale services.
        if (gdOpener) {
            accessory.removeService(gdOpener);
        }
        // Add the garage door opener service to the accessory.
        const gdService = new hap.Service.GarageDoorOpener(accessory.displayName);
        // The initial door state when we first startup. The bias functions will help us
        // figure out what to do if we're caught in a tweener state.
        const doorCurrentState = this.doorCurrentStateBias(doorInitialState);
        const doorTargetState = this.doorTargetStateBias(doorCurrentState);
        // Add all the events to our accessory so we can act on HomeKit actions. We also set the current and target door states
        // based on our saved state from previous sessions.
        accessory
            .addService(gdService)
            .setCharacteristic(hap.Characteristic.CurrentDoorState, doorCurrentState)
            .setCharacteristic(hap.Characteristic.TargetDoorState, doorTargetState)
            .getCharacteristic(hap.Characteristic.TargetDoorState)
            .on("set" /* SET */, this.setDoorState.bind(this));
        // Add all the events to our accessory so we can tell HomeKit our state.
        accessory
            .getService(hap.Service.GarageDoorOpener)
            .getCharacteristic(hap.Characteristic.CurrentDoorState)
            .on("get" /* GET */, this.getDoorState.bind(this));
        // Make sure we can detect obstructions.
        accessory
            .getService(hap.Service.GarageDoorOpener)
            .getCharacteristic(hap.Characteristic.ObstructionDetected)
            .on("get" /* GET */, this.getObstructed.bind(this));
        // Update the firmware revision for this device.
        // Fun fact: This firmware information is stored on the gateway not the opener.
        const gwParent = this.myQ.Devices.find((x) => x.serial_number === device.parent_device_id);
        let gwBrand = "Liftmaster";
        let gwProduct = "myQ";
        if ((_a = gwParent === null || gwParent === void 0 ? void 0 : gwParent.state) === null || _a === void 0 ? void 0 : _a.firmware_version) {
            const gwInfo = this.myQ.getHwInfo(gwParent.serial_number);
            accessory
                .getService(hap.Service.AccessoryInformation)
                .getCharacteristic(hap.Characteristic.FirmwareRevision).updateValue(gwParent.state.firmware_version);
            // If we're able to lookup hardware information, use it. getHwInfo returns an object containing
            // device type and brand information.
            gwProduct = gwInfo === null || gwInfo === void 0 ? void 0 : gwInfo.product;
            gwBrand = gwInfo === null || gwInfo === void 0 ? void 0 : gwInfo.brand;
        }
        // Update the manufacturer information for this device.
        accessory
            .getService(hap.Service.AccessoryInformation)
            .getCharacteristic(hap.Characteristic.Manufacturer).updateValue(gwBrand);
        // Update the model information for this device.
        accessory
            .getService(hap.Service.AccessoryInformation)
            .getCharacteristic(hap.Characteristic.Model).updateValue(gwProduct);
        // Update the serial number for this device.
        accessory
            .getService(hap.Service.AccessoryInformation)
            .getCharacteristic(hap.Characteristic.SerialNumber).updateValue(device.serial_number);
        // Set us up to report battery status, but only if it's supported by the device.
        if (!this.batteryDeviceSupport && (this.doorPositionSensorBatteryStatus() !== -1)) {
            const gdService = accessory.getService(hap.Service.GarageDoorOpener);
            // Verify we've already setup the garage door service before trying to configure it.
            if (gdService) {
                gdService
                    .getCharacteristic(hap.Characteristic.StatusLowBattery)
                    .on("get" /* GET */, (callback) => {
                    callback(null, this.doorPositionSensorBatteryStatus());
                });
                // We only want to configure this once, not on each update.
                // Not the most elegant solution, but it gets the job done.
                this.batteryDeviceSupport = true;
                this.log("%s: Battery status support enabled.", accessory.displayName);
            }
        }
    }
    // Return whether or not the garage door detects an obstruction.
    getObstructed(callback) {
        // For a refresh of the door status, but we're really unconcerned about what it returns here.
        this.doorStatus();
        // See if we have an obstruction to alert on.
        if (this.obstructionDetected) {
            this.log("%s: Obstruction detected.", this.accessory.displayName);
        }
        callback(null, this.obstructionDetected);
    }
    // Return garage door status.
    getDoorState(callback) {
        const doorState = this.doorStatus();
        if (doorState === -1) {
            callback(new Error("Unable to determine the current door state."));
        }
        else {
            callback(null, doorState);
        }
    }
    // Open or close the garage door.
    setDoorState(value, callback) {
        const myQState = this.doorStatus();
        const accessory = this.accessory;
        const hap = this.hap;
        if (myQState === -1) {
            callback(new Error("Unable to determine the current door state."));
            return;
        }
        // If we are already opening or closing the garage door, we error out. myQ doesn't appear to allow
        // interruptions to an open or close command that is currently executing - it must be allowed to
        // complete its action before accepting a new one.
        if ((myQState === hap.Characteristic.CurrentDoorState.OPENING) || (myQState === hap.Characteristic.CurrentDoorState.CLOSING)) {
            const actionExisting = myQState === hap.Characteristic.CurrentDoorState.OPENING ? "opening" : "closing";
            const actionAttempt = value === hap.Characteristic.TargetDoorState.CLOSED ? "close" : "open";
            this.log("%s: Unable to %s door while currently attempting to complete %s. myQ must complete it's existing action before attempting a new one.", accessory.displayName, actionAttempt, actionExisting);
            callback(new Error("Unable to accept a new set event while another is completing."));
        }
        else if (value === hap.Characteristic.TargetDoorState.CLOSED) {
            // HomeKit is informing us to close the door, but let's make sure it's not already closed first.
            if (myQState !== hap.Characteristic.CurrentDoorState.CLOSED) {
                // We set this to closing instead of closed for a couple of reasons. First, myQ won't immediately execute
                // this command for safety reasons - it enforces a warning tone for a few seconds before it starts the action.
                // Second, HomeKit gets confused with our multiple updates of this value, so we'll set it to closing and hope
                // for the best.
                accessory
                    .getService(hap.Service.GarageDoorOpener)
                    .getCharacteristic(hap.Characteristic.CurrentDoorState).updateValue(hap.Characteristic.CurrentDoorState.CLOSING);
                // Execute this command and begin polling myQ for state changes.
                this.doorCommand(hap.Characteristic.TargetDoorState.CLOSED);
            }
            callback(null);
        }
        else if (value === hap.Characteristic.TargetDoorState.OPEN) {
            // HomeKit is informing us to open the door, but we don't want to act if it's already open.
            if (myQState !== hap.Characteristic.CurrentDoorState.OPEN) {
                // We set this to opening instad of open because we want to show our state transitions to HomeKit and end users.
                accessory
                    .getService(hap.Service.GarageDoorOpener)
                    .getCharacteristic(hap.Characteristic.CurrentDoorState).updateValue(hap.Characteristic.CurrentDoorState.OPENING);
                // Execute this command and begin polling myQ for state changes.
                this.doorCommand(hap.Characteristic.TargetDoorState.OPEN);
            }
            callback(null);
        }
        else {
            // HomeKit has told us something that we don't know how to handle.
            this.log("%s: Unknown SET event received: %s.", accessory.displayName, value);
            callback(new Error("Unknown SET event received: " + value));
        }
    }
    // Update our HomeKit status.
    async updateState() {
        var _a, _b, _c, _d, _e, _f;
        const accessory = this.accessory;
        const hap = this.hap;
        const oldState = accessory.context.doorState;
        const myQState = this.doorStatus();
        // HomeKit state decoder ring.
        const myQStateMap = {
            [hap.Characteristic.CurrentDoorState.OPEN]: "Open",
            [hap.Characteristic.CurrentDoorState.CLOSED]: "Closed",
            [hap.Characteristic.CurrentDoorState.OPENING]: "Opening",
            [hap.Characteristic.CurrentDoorState.CLOSING]: "Closing",
            [hap.Characteristic.CurrentDoorState.STOPPED]: "Stopped",
            [settings_1.MYQ_OBSTRUCTED]: "Obstructed"
        };
        // If we can't get our status, we're probably not able to connect to the myQ API.
        if (myQState === -1) {
            this.log("%s: Unable to determine the current door state.", accessory.displayName);
            return false;
        }
        if (oldState !== myQState) {
            this.log("%s: %s.", accessory.displayName, myQStateMap[myQState]);
            // Update the state in HomeKit. Thanks to @dxdc for suggesting looking at using updateValue
            // here instead of the more intuitive setCharacteristic due to inevitable race conditions and
            // set loops that can occur in HomeKit if you aren't careful.
            accessory.context.doorState = myQState;
            (_b = (_a = accessory.getService(hap.Service.GarageDoorOpener)) === null || _a === void 0 ? void 0 : _a.getCharacteristic(hap.Characteristic.CurrentDoorState)) === null || _b === void 0 ? void 0 : _b.updateValue(myQState);
            // We are only going to update the target state if our current state is NOT stopped. If we are stopped,
            // we are at the target state by definition. Unfortunately, the iOS Home app doesn't seem to correctly
            // report a stopped state, although you can find it correctly reported in other HomeKit apps like Eve Home.
            if (myQState !== hap.Characteristic.CurrentDoorState.STOPPED) {
                const targetState = this.doorTargetStateBias(myQState);
                (_d = (_c = accessory.getService(hap.Service.GarageDoorOpener)) === null || _c === void 0 ? void 0 : _c.getCharacteristic(hap.Characteristic.TargetDoorState)) === null || _d === void 0 ? void 0 : _d.updateValue(targetState);
            }
            // When we detect any state change, we want to increase our polling resolution to provide timely updates.
            this.platform.configPoll.count = 0;
            this.platform.poll(0);
        }
        const batteryStatus = this.doorPositionSensorBatteryStatus();
        // Update battery status only if it's supported by the device.
        if (batteryStatus !== -1) {
            (_f = (_e = accessory.getService(hap.Service.GarageDoorOpener)) === null || _e === void 0 ? void 0 : _e.getCharacteristic(hap.Characteristic.StatusLowBattery)) === null || _f === void 0 ? void 0 : _f.updateValue(batteryStatus);
        }
        return true;
    }
    // Return the status of the door. This function maps myQ door status to HomeKit door status.
    doorStatus() {
        // Door state cheat sheet.
        // autoreverse is how the myQ API communicated an obstruction...go figure. Unfortunately, it
        // only seems to last the duration of the door reopening (reversal).
        const doorStates = {
            open: this.hap.Characteristic.CurrentDoorState.OPEN,
            closed: this.hap.Characteristic.CurrentDoorState.CLOSED,
            opening: this.hap.Characteristic.CurrentDoorState.OPENING,
            closing: this.hap.Characteristic.CurrentDoorState.CLOSING,
            stopped: this.hap.Characteristic.CurrentDoorState.STOPPED,
            autoreverse: settings_1.MYQ_OBSTRUCTED
        };
        const device = this.accessory.context.device;
        if (!device) {
            this.log("%s: Can't find the associated device in the myQ API.", this.accessory.displayName);
            return -1;
        }
        // Retrieve the door state from myQ and map it to HomeKit.
        const myQState = doorStates[device.state.door_state];
        if (myQState === undefined) {
            this.log("%s: Unknown door state encountered: %s.", this.accessory.displayName, device.state.door_state);
            return -1;
        }
        // Obstructed states in the myQ API remain active for a very small period of time. Furthermore, the way
        // HomeKit informs you of an obstructed state is through a status update on the Home app home screen.
        // This ultimately means that an obstructed state has a very small chance of actually being visible to
        // a user unless they happen to be looking at the Home app at the exact moment the obstruction is detected.
        // To ensure the user has a reasonable chance to notice the obstructed state, we will alert a user for up
        // to MYQ_OBSTRUCTION_ALERT_DURATION seconds after the last time we detected an obstruction before clearing
        // out the alert.
        if (myQState === settings_1.MYQ_OBSTRUCTED) {
            // Clear any other timer that might be out there for obstructions.
            clearTimeout(this.obstructionTimer);
            // Obstruction detected.
            this.obstructionDetected = true;
            const accessory = this.accessory;
            const hap = this.hap;
            const self = this;
            // Set the timer for clearing out the obstruction state.
            this.obstructionTimer = setTimeout(() => {
                self.obstructionDetected = false;
                accessory
                    .getService(hap.Service.GarageDoorOpener)
                    .getCharacteristic(hap.Characteristic.ObstructionDetected).updateValue(self.obstructionDetected);
                self.log("%s: Obstruction cleared.", self.accessory.displayName);
            }, settings_1.MYQ_OBSTRUCTION_ALERT_DURATION * 1000);
        }
        return myQState;
    }
    // Open or close the door for an accessory.
    doorCommand(command) {
        let myQCommand;
        // Translate the command from HomeKit to myQ.
        switch (command) {
            case this.hap.Characteristic.TargetDoorState.OPEN:
                myQCommand = "open";
                break;
            case this.hap.Characteristic.TargetDoorState.CLOSED:
                myQCommand = "close";
                break;
            default:
                this.log("%s: Unknown door command encountered: %s.", this.accessory.displayName, command);
                return false;
        }
        const device = this.accessory.context.device;
        if (!device) {
            this.log("%s: Can't find the associated device in the myQ API.", this.accessory.displayName);
            return false;
        }
        // Execute the command.
        this.myQ.execute(device.serial_number, myQCommand);
        // Increase the frequency of our polling for state updates to catch any updates from myQ.
        // This will trigger polling at activeRefreshInterval until activeRefreshDuration is hit. If you
        // query the myQ API too quickly, the API won't have had a chance to begin executing our command.
        this.platform.configPoll.count = 0;
        this.platform.poll(0);
        return true;
    }
    // Return our bias for what the current door state should be. This is primarily used for our initial bias on startup.
    doorCurrentStateBias(myQState) {
        // Our current state reflects having to take an opinion on what open or closed means to
        // HomeKit. For the obvious states, this is easy. For some of the edge cases, it can be less so.
        // Our north star is that if we are in an obstructed state, we are open.
        switch (myQState) {
            case this.hap.Characteristic.CurrentDoorState.OPEN:
            case this.hap.Characteristic.CurrentDoorState.OPENING:
            case settings_1.MYQ_OBSTRUCTED:
                return this.hap.Characteristic.CurrentDoorState.OPEN;
            case this.hap.Characteristic.CurrentDoorState.STOPPED:
                return this.hap.Characteristic.CurrentDoorState.STOPPED;
            case this.hap.Characteristic.CurrentDoorState.CLOSED:
            case this.hap.Characteristic.CurrentDoorState.CLOSING:
            default:
                return this.hap.Characteristic.CurrentDoorState.CLOSED;
        }
    }
    // Return our bias for what the target door state should be.
    doorTargetStateBias(myQState) {
        // We need to be careful with respect to the target state and we need to make some
        // reasonable assumptions about where we intend to end up. If we are opening or closing,
        // our target state needs to be the completion of those actions. If we're stopped or
        // obstructed, we're going to assume the desired target state is to be open, since that
        // is the typical garage door behavior, and it's impossible for us to know with reasonable
        // certainty what the original intention of the action was.
        switch (myQState) {
            case this.hap.Characteristic.CurrentDoorState.OPEN:
            case this.hap.Characteristic.CurrentDoorState.OPENING:
            case this.hap.Characteristic.CurrentDoorState.STOPPED:
            case settings_1.MYQ_OBSTRUCTED:
                return this.hap.Characteristic.TargetDoorState.OPEN;
            case this.hap.Characteristic.CurrentDoorState.CLOSED:
            case this.hap.Characteristic.CurrentDoorState.CLOSING:
            default:
                return this.hap.Characteristic.TargetDoorState.CLOSED;
        }
    }
    // Return the battery status of the door sensor, if supported on the device.
    doorPositionSensorBatteryStatus() {
        const device = this.accessory.context.device;
        if (!device) {
            this.log("%s: Can't find the associated device in the myQ API.", this.accessory.displayName, this.accessory.UUID);
            return -1;
        }
        // If we don't find the dps_low_battery_mode attribute, then this device may not support it.
        if (device.state.dps_low_battery_mode === undefined) {
            return -1;
        }
        return device.state.dps_low_battery_mode ? this.hap.Characteristic.StatusLowBattery.BATTERY_LEVEL_LOW :
            this.hap.Characteristic.StatusLowBattery.BATTERY_LEVEL_NORMAL;
    }
}
exports.myQGarageDoor = myQGarageDoor;
//# sourceMappingURL=myq-garagedoor.js.map