"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.myQPlatform = void 0;
const myq_api_1 = require("./myq-api");
const myq_garagedoor_1 = require("./myq-garagedoor");
const settings_1 = require("./settings");
const util_1 = __importDefault(require("util"));
let hap;
let Accessory;
class myQPlatform {
    constructor(log, config, api) {
        this.accessories = [];
        this.batteryDeviceSupport = {};
        this.configuredAccessories = {};
        this.unsupportedDevices = {};
        this.configOptions = [];
        this.configPoll = {
            refreshInterval: settings_1.MYQ_DEVICE_REFRESH_INTERVAL,
            activeRefreshInterval: settings_1.MYQ_ACTIVE_DEVICE_REFRESH_INTERVAL,
            activeRefreshDuration: settings_1.MYQ_ACTIVE_DEVICE_REFRESH_DURATION,
            maxCount: 0,
            count: 0
        };
        this.api = api;
        this.log = log;
        Accessory = api.platformAccessory;
        hap = api.hap;
        // We can't start without being configured.
        if (!config) {
            return;
        }
        // We need login credentials or we're not starting.
        if (!config.email || !config.password) {
            this.log("No myQ login credentials configured.");
            return;
        }
        // Check configuration parameters.
        this.debugMode = config.debug === true;
        this.debug("Debug logging on. Expect a lot of data.");
        if (config.activeRefreshDuration) {
            this.configPoll.activeRefreshDuration = config.activeRefreshDuration;
        }
        if (config.activeRefreshInterval) {
            this.configPoll.activeRefreshInterval = config.activeRefreshInterval;
        }
        if (config.options) {
            this.configOptions = config.options;
        }
        if (config.refreshInterval) {
            this.configPoll.refreshInterval = config.refreshInterval;
        }
        this.configPoll.maxCount = this.configPoll.activeRefreshDuration / this.configPoll.activeRefreshInterval;
        this.configPoll.count = this.configPoll.maxCount;
        // Initialize our connection to the myQ API.
        this.myQ = new myq_api_1.myQApi(this, config.email, config.password, config.appId);
        // Avoid a prospective race condition by waiting to begin our polling until Homebridge is done
        // loading all the cached accessories it knows about, and calling configureAccessory() on each.
        //
        // Fire off our polling, with an immediate status refresh to begin with to provide us that responsive feeling.
        api.on("didFinishLaunching" /* DID_FINISH_LAUNCHING */, this.poll.bind(this, this.configPoll.refreshInterval * -1));
    }
    // This gets called when homebridge restores cached accessories at startup. We
    // intentionally avoid doing anything significant here, and save all that logic
    // for device discovery.
    configureAccessory(accessory) {
        // Zero out the myQ device pointer on startup. This will be set by device discovery.
        accessory.context.device = null;
        // Add this to the accessory array so we can track it.
        this.accessories.push(accessory);
    }
    // Discover new myQ devices and sync existing ones with the myQ API.
    discoverAndSyncAccessories() {
        // Remove any device objects from now-stale accessories.
        for (const accessory of this.accessories) {
            // We only need to do this if the device object is set.
            if (!accessory.context.device) {
                continue;
            }
            // Check to see if this accessory's device object is still in myQ or not.
            if (!this.myQ.Devices.some((x) => x.serial_number === accessory.context.device.serial_number)) {
                accessory.context.device = null;
            }
        }
        // Iterate through the list of devices that myQ has returned and sync them with what we show HomeKit.
        for (const device of this.myQ.Devices) {
            // If we have no serial number or device family, something is wrong.
            if (!device.serial_number || !device.device_family) {
                continue;
            }
            // We are only interested in garage door openers. Perhaps more types in the future.
            if (device.device_family.indexOf("garagedoor") === -1) {
                // Unless we are debugging device discovery, ignore any gateways.
                // These are typically gateways, hubs, etc. that shouldn't be causing us to alert anyway.
                if (!this.debugMode && device.device_family === "gateway") {
                    continue;
                }
                // If we've already informed the user about this one, we're done.
                if (this.unsupportedDevices[device.serial_number]) {
                    continue;
                }
                // Notify the user we see this device, but we aren't adding it to HomeKit.
                this.unsupportedDevices[device.serial_number] = true;
                this.log("myQ device family '%s' is not currently supported, ignoring: %s.", device.device_family, this.myQ.getDeviceName(device));
                continue;
            }
            // Exclude or include certain openers based on configuration parameters.
            if (!this.deviceVisible(device)) {
                continue;
            }
            // Generate this device's unique identifier.
            const uuid = hap.uuid.generate(device.serial_number);
            let accessory;
            // See if we already know about this accessory or if it's truly new. If it is new, add it to HomeKit.
            if ((accessory = this.accessories.find((x) => x.UUID === uuid)) === undefined) {
                accessory = new Accessory(device.name, uuid);
                this.log("%s: Adding %s device to HomeKit: %s.", device.name, device.device_family, this.myQ.getDeviceName(device));
                // Register this accessory with homebridge and add it to the accessory array so we can track it.
                this.api.registerPlatformAccessories(settings_1.PLUGIN_NAME, settings_1.PLATFORM_NAME, [accessory]);
                this.accessories.push(accessory);
            }
            // Link the accessory to it's device object.
            accessory.context.device = device;
            // Setup the myQ device if it hasn't been configured yet.
            if (!this.configuredAccessories[accessory.UUID]) {
                // Eventually switch on multiple types of myQ devices. For now, it's garage doors only...
                this.configuredAccessories[accessory.UUID] = new myq_garagedoor_1.myQGarageDoor(this, accessory);
                // Refresh the accessory cache with these values.
                this.api.updatePlatformAccessories([accessory]);
            }
        }
        // Remove myQ devices that are no longer found in the myQ API, but we still have in HomeKit.
        for (const oldAccessory of this.accessories) {
            const device = oldAccessory.context.device;
            // We found this accessory in myQ. Figure out if we really want to see it in HomeKit.
            if (device && this.deviceVisible(device)) {
                continue;
            }
            this.log("%s: Removing myQ device from HomeKit.", oldAccessory.displayName);
            delete this.configuredAccessories[oldAccessory.UUID];
            this.accessories.splice(this.accessories.indexOf(oldAccessory), 1);
            this.api.unregisterPlatformAccessories(settings_1.PLUGIN_NAME, settings_1.PLATFORM_NAME, [oldAccessory]);
        }
        return true;
    }
    // Update HomeKit with the latest status from myQ.
    async updateAccessories() {
        // Refresh the full device list from the myQ API.
        if (!(await this.myQ.refreshDevices())) {
            return false;
        }
        // Sync myQ status and check for any new or removed accessories.
        await this.discoverAndSyncAccessories();
        // Iterate through our accessories and update its status with the corresponding myQ status.
        for (const key in this.configuredAccessories) {
            await this.configuredAccessories[key].updateState();
        }
        return true;
    }
    // Periodically poll the myQ API for status.
    poll(delay) {
        let refresh = this.configPoll.refreshInterval + delay;
        // Clear the last polling interval out.
        clearTimeout(this.pollingTimer);
        // Normally, count just increments on each call. However, when we want to increase our
        // polling frequency, count is set to 0 (elsewhere in the plugin) to put us in a more
        // frequent polling mode. This is determined by the values configured for
        // activeRefreshDuration and activeRefreshInterval which specify the maximum length of time
        // for this increased polling frequency (activeRefreshDuration) and the actual frequency of
        // each update (activeRefreshInterval).
        if (this.configPoll.count < this.configPoll.maxCount) {
            refresh = this.configPoll.activeRefreshInterval + delay;
            this.configPoll.count++;
        }
        // Setup periodic update with our polling interval.
        const self = this;
        this.pollingTimer = setTimeout(async () => {
            // Refresh our myQ information and gracefully handle myQ errors.
            if (!(await self.updateAccessories())) {
                self.configPoll.count = self.configPoll.maxCount - 1;
            }
            // Fire off the next polling interval.
            self.poll(0);
        }, refresh * 1000);
    }
    // Utility function to let us know if a myQ device should be visible in HomeKit or not.
    deviceVisible(device) {
        // There are a couple of ways to hide and show devices that we support. The rules of the road are:
        //
        // 1. Explicitly hiding, or showing a gateway device propogates to all the devices that are plugged
        //    into that gateway. So if you have multiple gateways but only want one exposed in this plugin,
        //    you may do so by hiding it.
        //
        // 2. Explicitly hiding, or showing an opener device by its serial number will always override the above.
        //    This means that it's possible to hide a gateway, and all the openers that are attached to it, and then
        //    override that behavior on a single opener device that it's connected to.
        //
        // Nothing configured - we show all myQ devices to HomeKit.
        if (!this.configOptions) {
            return true;
        }
        // No device. Sure, we'll show it.
        if (!device) {
            return true;
        }
        // We've explicitly enabled this opener.
        if (this.configOptions.indexOf("Show." + (device.serial_number)) !== -1) {
            return true;
        }
        // We've explicitly hidden this opener.
        if (this.configOptions.indexOf("Hide." + device.serial_number) !== -1) {
            return false;
        }
        // If we don't have a gateway device attached to this opener, we're done here.
        if (!device.parent_device_id) {
            return true;
        }
        // We've explicitly shown the gateway this opener is attached to.
        if (this.configOptions.indexOf("Show." + device.parent_device_id) !== -1) {
            return true;
        }
        // We've explicitly hidden the gateway this opener is attached to.
        if (this.configOptions.indexOf("Hide." + device.parent_device_id) !== -1) {
            return false;
        }
        // Nothing special to do - make this opener visible.
        return true;
    }
    // Utility for debug logging.
    debug(message, ...parameters) {
        if (this.debugMode) {
            this.log(util_1.default.format(message, ...parameters));
        }
    }
}
exports.myQPlatform = myQPlatform;
//# sourceMappingURL=myq-platform.js.map